///--------------------------------------------------------------------------------------
// BeSpecific.cpp
//
// By Patrick Edson (pedson@bu.edu). Created April 1998.
//
// Isolates the BeOS specific code from the SpriteWorld code as much as possible.  Our
// system is simple: create a window with single view to draw the spriteworld in and
// spawn a thread to run the animation.
///--------------------------------------------------------------------------------------

#include <SWIncludes.h>
#include <SWFPSReport.h>

#include "BeSpecific.h"
#include "Simple.h"

/************************/
/* BeOS related Globals */
/************************/

thread_id		gAnimThread;				// thread id of our animation thread
BPath			*gAppPath;					// application path so we can find our sprites
SimpleWindow	*gWindowP;					
SimpleView		*gViewP;					// a simple view to hold out spriteworld

/*******************************/
/* SpriteWorld related Globals */
/*******************************/

extern SpriteWorldPtr	gSpriteWorldP;
extern SpriteLayerPtr	gSpriteLayerP;
extern SpritePtr		gMasterBallSpriteP;
extern uint32			gFrames;

///--------------------------------------------------------------------------------------
// main
///--------------------------------------------------------------------------------------
main()
{
	SimpleApp		theApp;
	
	srand((long) system_time());
	
	theApp.Run();
	
	return (0);
}

///--------------------------------------------------------------------------------------
// SimpleApp
//
// Create the window and get the application directory.
//
///--------------------------------------------------------------------------------------
SimpleApp::SimpleApp() : BApplication("application/x-vnd.Simple")
{
	BRect	frameRect;
	
	SetHomeDirectory();
	
	frameRect.Set(0, 0, kWindowWidth, kWindowHeight);
	frameRect.OffsetBy(kWindowOffset, kWindowOffset);
		
	gWindowP = new SimpleWindow( frameRect );

	frameRect.OffsetTo(B_ORIGIN);
	
	gViewP = new SimpleView(frameRect);

	SetUpSpriteWorld();
	CreateBallSprite();
	AddSprites();
	
	gWindowP->AddChild(gSpriteWorldP->windowFrameP->framePort);
		
	SWUpdateSpriteWorld(gSpriteWorldP, true);
	
	gAnimThread = spawn_thread(RunAnimation, "Animation_Thread",
						B_NORMAL_PRIORITY, gSpriteWorldP);
						
	resume_thread(gAnimThread);
}

///--------------------------------------------------------------------------------------
// SetHomeDirectory
//
// Get the BPath for the application directory so we can find the sprites.
//
///--------------------------------------------------------------------------------------
void
SimpleApp::SetHomeDirectory()
{
	fLaunchDirectory = new BPath;
	
	app_info info; 
	GetAppInfo(&info); 

	BEntry appentry; 
	appentry.SetTo(&info.ref); 

	appentry.GetPath(fLaunchDirectory);
	
	fLaunchDirectory->GetParent(fLaunchDirectory);
	
	gAppPath = fLaunchDirectory;
}

#pragma mark -

///--------------------------------------------------------------------------------------
// SimpleWindow
//
// Create the one child view of the window for the spriteworld.
//
///--------------------------------------------------------------------------------------
SimpleWindow::SimpleWindow(BRect frame)
			: BWindow(frame, "Simple Window", B_TITLED_WINDOW, B_NOT_RESIZABLE)
{
	Show();
}


///--------------------------------------------------------------------------------------
// QuitRequested
//
// Override to quit app.
//
///--------------------------------------------------------------------------------------
bool
SimpleWindow::QuitRequested()
{
	be_app->ShowCursor();

	ShowResults(gFrames);	// Show FPS report

	be_app->PostMessage(B_QUIT_REQUESTED);
	
	return true;
}

#pragma mark -
///--------------------------------------------------------------------------------------
// SimpleView
///--------------------------------------------------------------------------------------
SimpleView::SimpleView(BRect frame)
			: BView(frame, "Simple_View", B_FOLLOW_NONE, B_WILL_DRAW | B_PULSE_NEEDED)
{
	SetDrawingMode(B_OP_OVER);
}

///--------------------------------------------------------------------------------------
// Draw
//
// Refresh the spriteworld if the window needs updating.
//
///--------------------------------------------------------------------------------------
void
SimpleView::Draw(BRect updateRect)
{
	SWUpdateSpriteWorld(gSpriteWorldP, true);
}

///--------------------------------------------------------------------------------------
// MouseDown
//
// Override to quit w/mouse down event as the original does.
//
///--------------------------------------------------------------------------------------
void
SimpleView::MouseDown(BPoint point)
{
	Looper()->PostMessage(B_QUIT_REQUESTED);
}

///--------------------------------------------------------------------------------------
// MouseMoved
//
// Override to hide the cursor when it is in the spriteworld view (and to show it when
// it leaves).
//
///--------------------------------------------------------------------------------------
void
SimpleView::MouseMoved(BPoint point, uint32 transit, const BMessage *message)
{
	switch(transit)
	{
		case B_ENTERED_VIEW:
			be_app->HideCursor();
		break;
		
		case B_EXITED_VIEW:
			be_app->ShowCursor();
		break;
	
		default:
			BView::MouseMoved(point, transit, message);
	}
}
