///--------------------------------------------------------------------------------------
// Simple.cpp
///--------------------------------------------------------------------------------------

#include <SWGameUtils.h>
#include <SWFPSReport.h>
#include <SWApplication.h>

#include "BeSpecific.h"
#include "Simple.h"

#define kWorldRectInset			0			// Makes the SpriteWorld smaller than the window.
#define kMaxFPS					0			// Keep the animation from going too fast

#define kSpriteWidth			44
#define kSpriteHeight			44
#define kSpriteVertOffset		0
#define kSpriteHorizOffset		0
#define kSpriteFrames			1
#define kNumSprites				10
#define kMaxSpriteMoveDelta		8

/*******************************/
/* SpriteWorld related Globals */
/*******************************/

SpriteWorldPtr	gSpriteWorldP;
SpriteLayerPtr	gSpriteLayerP;
SpritePtr		gMasterBallSpriteP;
uint32			gFrames;

/************************/
/* BeOS related Globals */
/************************/

extern BPath			*gAppPath;
extern SimpleWindow		*gWindowP;
extern SimpleView		*gViewP;

///--------------------------------------------------------------------------------------
//	SetUpSpriteWorld
///--------------------------------------------------------------------------------------
void
SetUpSpriteWorld(void)
{
	OSErr			err;
	Rect			worldRect;

		// Initialize the SpriteWorld package
	err = SWEnterSpriteWorld();
	FatalError(err);
	
	worldRect = gWindowP->Bounds();
	worldRect.InsetBy(kWorldRectInset, kWorldRectInset);
	
		// Create the SpriteWorld
	err = SWCreateSpriteWorldFromWindow(&gSpriteWorldP, gWindowP, gViewP, &worldRect, NULL, 32);
	FatalError(err);

		// Create the Sprite Layer
	err = SWCreateSpriteLayer(&gSpriteLayerP);
	FatalError(err);
	
		// Put the pieces together and lock the SpriteWorld. 
		// Note: since we are locking the SpriteWorld before adding the sprites,
		// we must make sure to lock each sprite individually when creating them.
	SWAddSpriteLayer(gSpriteWorldP, gSpriteLayerP);
	SWLockSpriteWorld(gSpriteWorldP);	

		// Draw the background
	DrawPatternInBackground();
		
		// Set up other miscellaneous options
	SWSetSpriteWorldMaxFPS(gSpriteWorldP, kMaxFPS);
}

///--------------------------------------------------------------------------------------
//	CreateBallSprite - load the master ball sprite and prepare it for cloning later
///--------------------------------------------------------------------------------------
void
CreateBallSprite(void)
{
	OSErr	err;
	
	BPath	*spritePath = new BPath;
	
	spritePath->SetTo(gAppPath->Path(), "ball.tiff");
	
	err = SWCreateSpriteFromSingleFileXY(gSpriteWorldP, &gMasterBallSpriteP, NULL,
		spritePath->Path(), spritePath->Path(), kSpriteWidth, kSpriteHeight, kSpriteHorizOffset,
		kSpriteVertOffset, kSpriteFrames, kFatMask);
	FatalError(err);

		// Set up the sprite (remember that it must be locked!)
	SWLockSprite(gMasterBallSpriteP);
	SWSetSpriteMoveBounds(gMasterBallSpriteP, &gSpriteWorldP->backRect);
	SWSetSpriteMoveProc(gMasterBallSpriteP, BallSpriteMoveProc);

}

///--------------------------------------------------------------------------------------
//	AddSprites - clone the master sprite, and add the clones to the SpriteWorld
///--------------------------------------------------------------------------------------
void
AddSprites(void)
{
	SpritePtr	newSpriteP;
	short		spriteNum;
	OSErr		err;

	for (spriteNum = 0; spriteNum < kNumSprites; spriteNum++)
	{
			// Clone the master sprite. The clone doesn't need to be locked,
			// since the master sprite was already locked.
		err = SWCloneSprite(gMasterBallSpriteP, &newSpriteP, NULL);
		FatalError(err);
		
		SWAddSprite(gSpriteLayerP, newSpriteP);
		
		SWSetSpriteLocation(newSpriteP,
				GetRandom(0, gSpriteWorldP->backRect.right-44), 
				GetRandom(0, gSpriteWorldP->backRect.bottom-44) );
		
		do
		{
			SWSetSpriteMoveDelta(newSpriteP,
					GetRandom(-kMaxSpriteMoveDelta, kMaxSpriteMoveDelta), 
					GetRandom(-kMaxSpriteMoveDelta, kMaxSpriteMoveDelta));
		} while (newSpriteP->horizMoveDelta == 0 || newSpriteP->vertMoveDelta == 0);
	}

}

//--------------------------------------------------------------------------------------
// RunAnimation
//--------------------------------------------------------------------------------------
int32
RunAnimation(void *arg)
{
	gFrames = 0;
	
	SpriteWorldPtr	spriteWorld = (SpriteWorldPtr)arg;

	StartTimer();	// For FPS report later
	
	while(true)
	{
		SWProcessSpriteWorld(spriteWorld);
		SWAnimateSpriteWorld(spriteWorld);

		if (gSpriteWorldP->frameHasOccurred)
			gFrames++;		
	}
}

///--------------------------------------------------------------------------------------
//  CleanUp
///--------------------------------------------------------------------------------------
void
CleanUp(void)
{
		// Dispose the master sprite, since it isn't included in the SpriteWorld
	SWDisposeSprite(gMasterBallSpriteP);
	
		// Dispose the SpriteWorld, including all sprites and layers currently in it
	SWDisposeSpriteWorld(gSpriteWorldP);
	SWExitSpriteWorld();

}

///--------------------------------------------------------------------------------------
//	DrawPatternInBackground
///--------------------------------------------------------------------------------------
void
DrawPatternInBackground()
{
	BRect		patternRect;
	BRect		windowRect;
	
	uint16		x, y, patternSize;
	
	windowRect = gSpriteWorldP->backRect;
	
	BPath	*spritePath = new BPath;
	spritePath->SetTo(gAppPath->Path(), "pattern.tiff");
	
	BBitmap *theMap = BTranslationUtils::GetBitmapFile(spritePath->Path());
	
	patternRect = theMap->Bounds();
	
	patternSize = patternRect.right - patternRect.left + 1;
	
	BRect		viewRect(0, 0, patternSize - 1, patternSize - 1);

	SWLockWindowFrame(gSpriteWorldP->backFrameP);
	
	for ( x = 0; x < windowRect.right; x += patternSize)
	{
		for ( y = 0; y < windowRect.bottom; y += patternSize)
		{
			gSpriteWorldP->backFrameP->framePort->DrawBitmap(theMap, patternRect, viewRect);
			
			viewRect.top	+= patternSize;
			viewRect.bottom	+= patternSize;
		}
		
		viewRect.left	+= patternSize; 
		viewRect.right	+= patternSize;
		
		viewRect.top	= 0;
		viewRect.bottom	= patternSize - 1;
	}
	
	gSpriteWorldP->backFrameP->framePort->Sync();
	SWUnlockWindowFrame(gSpriteWorldP->backFrameP);

}

///--------------------------------------------------------------------------------------
//  BallSpriteMoveProc
///--------------------------------------------------------------------------------------
SW_FUNC void BallSpriteMoveProc(SpritePtr ballSpriteP)
{	
	SWOffsetSprite(ballSpriteP, ballSpriteP->horizMoveDelta, ballSpriteP->vertMoveDelta);
	
	(void)SWBounceSprite(ballSpriteP);
	//(void)SWWrapSprite(ballSpriteP);
}
