/*  -----------------------------------------------------------------------------------
    -----------------------------------------------------------------------------------
	BlitPixieScaled8Bit.c
	
    by Brian Roddy
    
    5/31/97

	These routines are 8-bit and 16-bit blitters that dynamically shrink or 
	stretch the sprite.  It uses the srcRect and dstRect to figure out the amount 
	of scaling to do.  
	
	Use SWSetSpriteScaledSize to set the scaling size.  It takes a sprite pointer and
	a width and a height.  e.g.:
	
 	SWSetSpriteScaledSize(mySpritePtr, 20, 50);  // Scale it
	
	Use a width or height <= 0 to turn off scaling.  Note that just resetting 
	the sprite's draw procedure will not turn off scaling properly.  Make sure to 
	call SWSetSpriteScaledSize with a negative width or height.  e.g.:

 	SWSetSpriteScaledSize(mySpritePtr, -1, -1);  // Reset it back to normal
 	
	Also note that scaling is done while drawing to the screen.  As a result, it can 
	the pixels in the frames don't match what shows up on the screen.  This means that 
	functions which rely on the sprite's mask, such as SWPixelCollision, won't work as 
	expected.  Thus using the pixel or mask collisions routines with scaleded sprites 
	won't work accurately.
	
	-----------------------------------------------------------------------------------
	Implementation Notes:

	These blitters use the "Bresenham line drawing algorithm" to do the scaling.  This is
	the standard technique used for many scaling and texture mapping algorithms.  There
	are additional ways to speed this up by precomputing the scaling indexes and storing
	them in an array.  (It's still pretty fast as is.)

    This source code is available for free use.

    ----------------------------------------------------------------------------------- 
    ----------------------------------------------------------------------------------- */


#include "BlitPixieScaled.h"

#pragma mark ----------- Public API -------------
SW_FUNC OSErr SWSetSpriteScaledSize(
	SpritePtr srcSpriteP,
	long width,
	long height) 
{
	short 		currentPixelDepth;
	OSErr		err;
	
	
	currentPixelDepth = (*srcSpriteP->frameArray[0]->framePort->portPixMap)->pixelSize;
	err = kWrongDepthErr;
	
	if (currentPixelDepth == 8)
	{
		if ((width <= 0) || (height <= 0)) {
			srcSpriteP->scaledFrameRect.left = -1;
			srcSpriteP->scaledFrameRect.top = -1;
			srcSpriteP->scaledFrameRect.right = -1;
			srcSpriteP->scaledFrameRect.bottom = -1;
			srcSpriteP->frameDrawProc = BlitPixie8BitMaskDrawProc;
		} else {
			srcSpriteP->scaledFrameRect.left = 0;
			srcSpriteP->scaledFrameRect.top = 0;
			srcSpriteP->scaledFrameRect.right = width;
			srcSpriteP->scaledFrameRect.bottom = height;
			srcSpriteP->frameDrawProc = BlitPixie8BitScaledMaskDrawProc;
		}
		err = noErr;
	}
	else if (currentPixelDepth == 16)
	{
		if ((width <= 0) || (height <= 0)) {
			srcSpriteP->scaledFrameRect.left = -1;
			srcSpriteP->scaledFrameRect.top = -1;
			srcSpriteP->scaledFrameRect.right = -1;
			srcSpriteP->scaledFrameRect.bottom = -1;
			srcSpriteP->frameDrawProc = BlitPixieAllBitMaskDrawProc;
		} else {
			srcSpriteP->scaledFrameRect.left = 0;
			srcSpriteP->scaledFrameRect.top = 0;
			srcSpriteP->scaledFrameRect.right = width;
			srcSpriteP->scaledFrameRect.bottom = height;
			srcSpriteP->frameDrawProc = BlitPixie16BitScaledMaskDrawProc;
		}
		err = noErr;
	}
	SWSetStickyIfError( err );
	return err;
}

SW_FUNC int SWGetSpriteScaledWidth(
	SpritePtr srcSpriteP) {
	return srcSpriteP->scaledFrameRect.right - srcSpriteP->scaledFrameRect.left;
}

SW_FUNC int SWGetSpriteScaledHeight(
	SpritePtr srcSpriteP) {
	return srcSpriteP->scaledFrameRect.bottom - srcSpriteP->scaledFrameRect.top;
}




///--------------------------------------------------------------------------------------
///--------------------------------------------------------------------------------------
/// BLITTERS

extern char 				gSWmmuMode;
extern SpritePtr			gCurrentSpriteBeingDrawn;

#pragma mark ----------- 8 Bit Blitters -------------
///--------------------------------------------------------------------------------------
///--------------------------------------------------------------------------------------
/// 8-Bit
///--------------------------------------------------------------------------------------
///--------------------------------------------------------------------------------------

///--------------------------------------------------------------------------------------
//		BlitPixie8BitScaledRectDrawProc
///--------------------------------------------------------------------------------------

SW_FUNC void BlitPixie8BitScaledRectDrawProc(
	FramePtr srcFrameP,
	FramePtr dstFrameP,
	Rect* srcRect,
	Rect* dstRect)
{
	Rect 					srcBlitRect = *srcRect;
	Rect 					dstBlitRect = *dstRect;
	SInt8 					mmuMode;
	int srcNumBytesPerRow = srcBlitRect.right - srcBlitRect.left;

	CLIP_RECT(dstFrameP->frameRect, srcBlitRect, dstBlitRect);
	START_MMU_MODE();
	

	BlitPixie8BitScaled(
			// calculate the address of the first byte of the source
		(unsigned char *)(srcFrameP->frameBaseAddr + 
			(srcFrameP->scanLinePtrArray[srcBlitRect.top - srcFrameP->frameRect.top]) + 
			srcBlitRect.left),

			// calculate the address of the first byte of the destination
		(unsigned char *)(dstFrameP->frameBaseAddr + 
			(dstFrameP->scanLinePtrArray[dstBlitRect.top]) + 
			dstBlitRect.left),

			// calculate the number of rows to blit
		(srcBlitRect.bottom - srcBlitRect.top),
		(dstBlitRect.bottom - dstBlitRect.top),

			// number of bytes in source and dest rows
		srcNumBytesPerRow,
		(dstBlitRect.right - dstBlitRect.left),

		// for PPC, just frameRowBytes
		srcFrameP->frameRowBytes,
		dstFrameP->frameRowBytes
		);

	STOP_MMU_MODE();
}

///--------------------------------------------------------------------------------------
//		BlitPixie8BitScaledMaskDrawProc
///--------------------------------------------------------------------------------------

SW_FUNC void BlitPixie8BitScaledMaskDrawProc(
	FramePtr srcFrameP,
	FramePtr dstFrameP,
	Rect *srcRect,
	Rect *dstRect)
{
	Rect dstBlitRect = *dstRect;
	Rect srcBlitRect = *srcRect;
	unsigned long 		srcBaseOffset;
	SInt8 				mmuMode;
	int srcNumBytesPerRow = srcBlitRect.right - srcBlitRect.left;
	
	//CLIP_RECT(dstFrameP->frameRect, srcBlitRect, dstBlitRect);	
	START_MMU_MODE();

	// calculate the offset to the first byte of the source
	srcBaseOffset = srcFrameP->scanLinePtrArray[srcBlitRect.top - 
		srcFrameP->frameRect.top] + srcBlitRect.left;

	BlitPixieMask8BitScaled(
			// calculate the address of the first byte of the source
		(unsigned char *)(srcFrameP->frameBaseAddr + srcBaseOffset),

			// calculate the address of the first byte of the destination
		(unsigned char *)(dstFrameP->frameBaseAddr + 
			(dstFrameP->scanLinePtrArray[dstBlitRect.top]) + 
			dstBlitRect.left),

				// calculate the address of the first byte of the mask
		(unsigned char *)(srcFrameP->maskBaseAddr + srcBaseOffset),

			// calculate the number of rows to blit
		(srcBlitRect.bottom - srcBlitRect.top),
		(dstBlitRect.bottom - dstBlitRect.top),

			// number of bytes in source and dest rows
		srcNumBytesPerRow,
		(dstBlitRect.right - dstBlitRect.left),
		
		srcFrameP->frameRowBytes,
		dstFrameP->frameRowBytes
		);

	STOP_MMU_MODE();
}


///--------------------------------------------------------------------------------------
//		BlitPixie8BitScaledPartialMaskDrawProc
///--------------------------------------------------------------------------------------
SW_FUNC void BlitPixie8BitScaledPartialMaskDrawProc(
	FramePtr srcFrameP,
	FramePtr dstFrameP,
	Rect *srcRect,
	Rect *dstRect)
{
	// Same as above
	BlitPixie8BitScaledMaskDrawProc(srcFrameP, dstFrameP, srcRect, dstRect);

}

///--------------------------------------------------------------------------------------
//		BlitPixie8BitScaled
///--------------------------------------------------------------------------------------
void BlitPixie8BitScaled(
	unsigned char *srcPixelP,
	unsigned char *dstPixelP,
	register long srcRowsToCopy,
	register long dstRowsToCopy,
	register long srcNumBytesPerRow,
	register long dstNumBytesPerRow,
	register unsigned long srcOffset,
	register unsigned long dstOffset)
{
	register unsigned char *startSrcPixelP;
	register unsigned char *startDstPixelP;
	long verticalEps = 0;
	int curRow;
	
	startSrcPixelP = srcPixelP;
	startDstPixelP = dstPixelP;
	

	for (curRow = 0; curRow < dstRowsToCopy; curRow++)
	{
		int eps;
		srcPixelP = startSrcPixelP;
		dstPixelP = startDstPixelP;

		eps = 0;
		if (srcNumBytesPerRow >= dstNumBytesPerRow) {
	    	/* Shrink It */
	    	/* For now we don't do any pixel blending for this case */
			int srcIndex;
			int destIndex = 0;
			// Bresenham line draw
			for (srcIndex = 0; srcIndex < srcNumBytesPerRow; srcIndex++)  {
				eps += dstNumBytesPerRow;
				if ( (eps << 1) > srcNumBytesPerRow )  {
					dstPixelP[destIndex] = srcPixelP[srcIndex];
					destIndex++;
					eps -= srcNumBytesPerRow;
				}
			}
		} else {
	    	/* Stretch It */
			/* For now we don't do any anti aliasing for this case */
			int destIndex;
			int srcIndex = 0;
			for (destIndex = 0; destIndex < dstNumBytesPerRow; destIndex++)  {
				eps += srcNumBytesPerRow;
				dstPixelP[destIndex] = srcPixelP[srcIndex];
				if ( (eps << 1) > dstNumBytesPerRow )  {
					srcIndex++;
					eps -= dstNumBytesPerRow;
				}
			}
		}
		
			// bump to next row	
		startDstPixelP += dstOffset;

		if (srcRowsToCopy >= dstRowsToCopy) {
			verticalEps += dstRowsToCopy;
			if ((verticalEps << 1) > srcRowsToCopy)  {
				startSrcPixelP += srcOffset;
				verticalEps -= srcRowsToCopy;
			}
		} else {
			verticalEps += srcRowsToCopy;
			if ((verticalEps << 1) > dstRowsToCopy)  {
				startSrcPixelP += srcOffset;
				verticalEps -= dstRowsToCopy;
			}
		}
	}
}

///--------------------------------------------------------------------------------------
//		BlitPixieMask8BitScaled
///--------------------------------------------------------------------------------------
void BlitPixieMask8BitScaled(
	unsigned char *srcPixelP,
	unsigned char *dstPixelP,
	unsigned char *maskPixelP,
	register long srcRowsToCopy,
	register long dstRowsToCopy,
	register long srcNumBytesPerRow,
	register long dstNumBytesPerRow,
	register unsigned long srcOffset,
	register unsigned long dstOffset)
{
	register unsigned char *startSrcPixelP;
	register unsigned char *startDstPixelP;
	register unsigned char *startMaskPixelP;
	long verticalEps = 0;
	int curRow;
	
	startSrcPixelP = srcPixelP;
	startDstPixelP = dstPixelP;
	startMaskPixelP = maskPixelP;

	
	for (curRow = 0; curRow < dstRowsToCopy; curRow++)
	{
		int eps;
		srcPixelP = startSrcPixelP;
		dstPixelP = startDstPixelP;
		maskPixelP = startMaskPixelP;

		eps = 0;
		if (srcNumBytesPerRow >= dstNumBytesPerRow) {
	    	/* Shrink It */
	    	/* For now we don't do any pixel blending for this case */
			int srcIndex;
			int destIndex = 0;
			// Bresenham line draw
			for (srcIndex = 0; srcIndex < srcNumBytesPerRow; srcIndex++)  {
				eps += dstNumBytesPerRow;
				if ((eps << 1) > srcNumBytesPerRow)  {
					if (maskPixelP[srcIndex] == 0)
						dstPixelP[destIndex] = srcPixelP[srcIndex];
					destIndex++;
					eps -= srcNumBytesPerRow;
				}
			}
		} else {
	    	/* Stretch It */
			/* For now we don't do any anti aliasing for this case */
			int destIndex;
			int srcIndex = 0;
			for (destIndex = 0; destIndex < dstNumBytesPerRow; destIndex++)  {
				eps += srcNumBytesPerRow;
				if (maskPixelP[srcIndex] == 0)
					dstPixelP[destIndex] = srcPixelP[srcIndex];
				if ((eps << 1) > dstNumBytesPerRow)  {
					srcIndex++;
					eps -= dstNumBytesPerRow;
				}
			}
		}
		
			// bump to next row	
		startDstPixelP += dstOffset;

		if (srcRowsToCopy >= dstRowsToCopy) {
			while ((verticalEps * 2) <= srcRowsToCopy) {
				// shrink
				startSrcPixelP += srcOffset;
				startMaskPixelP += srcOffset;
				verticalEps += dstRowsToCopy;
			}
			verticalEps -= srcRowsToCopy;

		} else {
			// grow
			verticalEps += srcRowsToCopy;
			if ((verticalEps * 2) > dstRowsToCopy)  {
				startSrcPixelP += srcOffset;
				startMaskPixelP += srcOffset;
				verticalEps -= dstRowsToCopy;
			}
		}
	}
}

///--------------------------------------------------------------------------------------
///--------------------------------------------------------------------------------------
///--------------------------------------------------------------------------------------
/// 16-Bit
///--------------------------------------------------------------------------------------
///--------------------------------------------------------------------------------------
///--------------------------------------------------------------------------------------
#pragma mark ----------- 16 Bit Blitters -------------

///--------------------------------------------------------------------------------------
//		BlitPixie16BitScaledRectDrawProc
///--------------------------------------------------------------------------------------

SW_FUNC void BlitPixie16BitScaledRectDrawProc(
	FramePtr srcFrameP,
	FramePtr dstFrameP,
	Rect* srcRect,
	Rect* dstRect)
{
	Rect 					srcBlitRect = *srcRect;
	Rect 					dstBlitRect = *dstRect;
	SInt8 					mmuMode;
	int srcNumShortsPerRow = srcBlitRect.right - srcBlitRect.left;

	CLIP_RECT(dstFrameP->frameRect, srcBlitRect, dstBlitRect);
	START_MMU_MODE();

	BlitPixie16BitScaled(
			// calculate the address of the first byte of the source
		(unsigned short *)(srcFrameP->frameBaseAddr + 
			(srcFrameP->scanLinePtrArray[srcBlitRect.top - srcFrameP->frameRect.top]) + 
			(srcBlitRect.left << 1)),

			// calculate the address of the first byte of the destination
		(unsigned short *)(dstFrameP->frameBaseAddr + 
			(dstFrameP->scanLinePtrArray[dstBlitRect.top]) + 
			(dstBlitRect.left << 1)),

			// calculate the number of rows to blit
		(srcBlitRect.bottom - srcBlitRect.top),
		(dstBlitRect.bottom - dstBlitRect.top),

			// number of shorts in source and dest rows
		srcNumShortsPerRow,
		(dstBlitRect.right - dstBlitRect.left),

		// for PPC, just frameRowBytes
		srcFrameP->frameRowBytes,
		dstFrameP->frameRowBytes
		);

	STOP_MMU_MODE();
}

///--------------------------------------------------------------------------------------
//		BlitPixie16BitScaledMaskDrawProc
///--------------------------------------------------------------------------------------

SW_FUNC void BlitPixie16BitScaledMaskDrawProc(
	FramePtr srcFrameP,
	FramePtr dstFrameP,
	Rect *srcRect,
	Rect *dstRect)
{
	Rect dstBlitRect = *dstRect;
	Rect srcBlitRect = *srcRect;
	unsigned long 		srcBaseOffset;
	SInt8 				mmuMode;
	int srcNumShortsPerRow = srcBlitRect.right - srcBlitRect.left;
	
	CLIP_RECT(dstFrameP->frameRect, srcBlitRect, dstBlitRect);	
	START_MMU_MODE();

	// calculate the offset to the first byte of the source
	srcBaseOffset = srcFrameP->scanLinePtrArray[srcBlitRect.top - 
		srcFrameP->frameRect.top] + (srcBlitRect.left << 1);

	BlitPixieMask16BitScaled(
			// calculate the address of the first byte of the source
		(unsigned short *)(srcFrameP->frameBaseAddr + srcBaseOffset),

			// calculate the address of the first byte of the destination
		(unsigned short *)(dstFrameP->frameBaseAddr + 
			(dstFrameP->scanLinePtrArray[dstBlitRect.top]) + 
			(dstBlitRect.left << 1)),

				// calculate the address of the first byte of the mask
		(unsigned short *)(srcFrameP->maskBaseAddr + srcBaseOffset),

			// calculate the number of rows to blit
		(srcBlitRect.bottom - srcBlitRect.top),
		(dstBlitRect.bottom - dstBlitRect.top),  
		
			// number of shorts in source and dest rows
		srcNumShortsPerRow,
		(dstBlitRect.right - dstBlitRect.left),  

		srcFrameP->frameRowBytes,
		dstFrameP->frameRowBytes
		);

	STOP_MMU_MODE();
}


///--------------------------------------------------------------------------------------
//		BlitPixie16BitScaledPartialMaskDrawProc
///--------------------------------------------------------------------------------------
SW_FUNC void BlitPixie16BitScaledPartialMaskDrawProc(
	FramePtr srcFrameP,
	FramePtr dstFrameP,
	Rect *srcRect,
	Rect *dstRect)
{
	// Same as above
	BlitPixie16BitScaledMaskDrawProc(srcFrameP, dstFrameP, srcRect, dstRect);

}

///--------------------------------------------------------------------------------------
//		BlitPixie16Bit
///--------------------------------------------------------------------------------------
void BlitPixie16BitScaled(
	unsigned short *srcPixelP,
	unsigned short *dstPixelP,
	register long srcRowsToCopy,
	register long dstRowsToCopy,
	register long srcNumShortsPerRow,
	register long dstNumShortsPerRow,
	register unsigned long srcOffset,
	register unsigned long dstOffset)
{
	register unsigned short *startSrcPixelP;
	register unsigned short *startDstPixelP;
	long verticalEps = 0;
	int curRow;
	
	startSrcPixelP = srcPixelP;
	startDstPixelP = dstPixelP;
	
	// divide by 2 to get number of shorts from bytes
	srcOffset >>= 1;
	dstOffset >>= 1;

	for (curRow = 0; curRow < dstRowsToCopy; curRow++)
	{
		int eps;
		srcPixelP = startSrcPixelP;
		dstPixelP = startDstPixelP;

		eps = 0;
		if (srcNumShortsPerRow >= dstNumShortsPerRow) {
	    	/* Shrink It */
	    	/* For now we don't do any pixel blending for this case */
			int srcIndex;
			int destIndex = 0;
			// Bresenham line draw
			for (srcIndex = 0; srcIndex < srcNumShortsPerRow; srcIndex++)  {
				eps += dstNumShortsPerRow;
				if ( (eps << 1) > srcNumShortsPerRow )  {
					dstPixelP[destIndex] = srcPixelP[srcIndex];
					destIndex++;
					eps -= srcNumShortsPerRow;
				}
			}
		} else {
	    	/* Stretch It */
			/* For now we don't do any anti aliasing for this case */
			int destIndex;
			int srcIndex = 0;
			for (destIndex = 0; destIndex < dstNumShortsPerRow; destIndex++)  {
				eps += srcNumShortsPerRow;
				dstPixelP[destIndex] = srcPixelP[srcIndex];
				if ( (eps << 1) > dstNumShortsPerRow )  {
					srcIndex++;
					eps -= dstNumShortsPerRow;
				}
			}
		}
		
			// bump to next row	
		startDstPixelP += dstOffset;

		if (srcRowsToCopy >= dstRowsToCopy) {
			verticalEps += dstRowsToCopy;
			if ((verticalEps << 1) > srcRowsToCopy)  {
				startSrcPixelP += srcOffset;
				verticalEps -= srcRowsToCopy;
			}
		} else {
			verticalEps += srcRowsToCopy;
			if ((verticalEps << 1) > dstRowsToCopy)  {
				startSrcPixelP += srcOffset;
				verticalEps -= dstRowsToCopy;
			}
		}
	}
}

///--------------------------------------------------------------------------------------
//		BlitPixieMask16BitScaled
///--------------------------------------------------------------------------------------
void BlitPixieMask16BitScaled(
	unsigned short *srcPixelP,
	unsigned short *dstPixelP,
	unsigned short *maskPixelP,
	register long srcRowsToCopy,
	register long dstRowsToCopy,
	register long srcNumShortsPerRow,
	register long dstNumShortsPerRow,
	register unsigned long srcOffset,
	register unsigned long dstOffset)
{
	register unsigned short *startSrcPixelP;
	register unsigned short *startDstPixelP;
	register unsigned short *startMaskPixelP;
	long verticalEps = 0;
	int curRow;
	
	startSrcPixelP = srcPixelP;
	startDstPixelP = dstPixelP;
	startMaskPixelP = maskPixelP;

	// divide by 2 to get number of shorts from bytes
	srcOffset >>= 1;
	dstOffset >>= 1;
	
	for (curRow = 0; curRow < dstRowsToCopy; curRow++)
	{
		int eps;
		srcPixelP = startSrcPixelP;
		dstPixelP = startDstPixelP;
		maskPixelP = startMaskPixelP;

		eps = 0;
		if (srcNumShortsPerRow >= dstNumShortsPerRow) {
	    	/* Shrink It */
	    	/* For now we don't do any pixel blending for this case */
			int srcIndex;
			int destIndex = 0;
			// Bresenham line draw
			for (srcIndex = 0; srcIndex < srcNumShortsPerRow; srcIndex++)  {
				eps += dstNumShortsPerRow;
				if ((eps << 1) > srcNumShortsPerRow)  {
					if (maskPixelP[srcIndex] == 0)
						dstPixelP[destIndex] = srcPixelP[srcIndex];
					destIndex++;
					eps -= srcNumShortsPerRow;
				}
			}
		} else {
	    	/* Stretch It */
			/* For now we don't do any anti aliasing for this case */
			int destIndex;
			int srcIndex = 0;
			for (destIndex = 0; destIndex < dstNumShortsPerRow; destIndex++)  {
				eps += srcNumShortsPerRow;
				if (maskPixelP[srcIndex] == 0)
					dstPixelP[destIndex] = srcPixelP[srcIndex];
				if ((eps << 1) > dstNumShortsPerRow)  {
					srcIndex++;
					eps -= dstNumShortsPerRow;
				}
			}
		}
		
			// bump to next row	
		startDstPixelP += dstOffset;

		if (srcRowsToCopy >= dstRowsToCopy) {
			while ((verticalEps * 2) <= srcRowsToCopy) {
				// shrink
				startSrcPixelP += srcOffset;
				startMaskPixelP += srcOffset;
				verticalEps += dstRowsToCopy;
			}
			verticalEps -= srcRowsToCopy;

		} else {
			// grow
			verticalEps += srcRowsToCopy;
			if ((verticalEps * 2) > dstRowsToCopy)  {
				startSrcPixelP += srcOffset;
				startMaskPixelP += srcOffset;
				verticalEps -= dstRowsToCopy;
			}
		}
	}
}


