///--------------------------------------------------------------------------------------
//	SWFastLine.c
///--------------------------------------------------------------------------------------

#include "SWFastLine.h"


/*
	SWDrawLine is an assembly language routine that will draw a line very quickly; much
	faster than QuickDraw. This can be useful for screen wipes or other special effects.
	It was originally written by Oleg Doperchuk, and is believed to be in the public
	domain. Minor modifications have been made to make it compatible with SpriteWorld.
	
	Currently, this file compiles only under Think C. If you can figure out how to get
	it to compile under CodeWarrior, please send Vern a copy with the modifications you
	have made.
	
	In order to use this function, you must be in 8-bit mode (256 colors or grays).
	Also remember to add this line to the beginning of your source code:
	#include "SWFastLine.h"
	
	The first parameter is a FramePtr indicating the frame in which you want the line
	to be drawn. This can be the SpriteWorld's windowFrameP, workFrameP, or backFrameP,
	or a sprite's framePtr.
	
	The srcPoint and dstPoint indicate the beginning and end of the line. No clipping
	is done, so it is up to you to make sure that the points contained in srcPoint
	and dstPoint are within the boundaries of the frame in which you are drawing.
	
	The color parameter is a numeric value between 0 and 255 indicating the color
	index value of the line. If you use the system palette, 0 will be white and 255
	is black, and every value in between is a different color. 
	
	Below is a short example that draws random lines in the window of the gSpriteWorldP.
	To try it out, you can substitute this function for the normal RunAnimation function
	you will find in the scrolling or tiling demos. Remember to #include "SWFastLine.h".
	
void	RunAnimation( void )
{
	Rect		windRect = gSpriteWorldP->windowFrameP->frameRect;
	Point		srcPoint, dstPoint;
	short		color;
	
	while (!Button())
	{
		srcPoint.h = GetRandom(windRect.left, windRect.right-1);
		srcPoint.v = GetRandom(windRect.top, windRect.bottom-1);
		dstPoint.h = GetRandom(windRect.left, windRect.right-1);
		dstPoint.v = GetRandom(windRect.top, windRect.bottom-1);
		color = GetRandom(0, 254);
		
		SWDrawLine(gSpriteWorldP->windowFrameP, srcPoint, dstPoint, color);
	}
}
*/


///--------------------------------------------------------------------------------------
//	SWDrawLine
///--------------------------------------------------------------------------------------

SW_FUNC void SWDrawLine(FramePtr dstFrameP, Point srcPoint, Point dstPoint, Byte color)
{
	Ptr				baseAddr = dstFrameP->frameBaseAddr + dstFrameP->scanLinePtrArray[0];
	unsigned long	rowBytes = dstFrameP->frameRowBytes;
	short			x1 = srcPoint.h, y1 = srcPoint.v, 
					x2 = dstPoint.h, y2 = dstPoint.v;
	SInt8 			mmuMode;
	
	if (gSWmmuMode != true32b)
	{
		mmuMode = true32b;
		SwapMMUMode(&mmuMode);
	}


	SW_ASM_BEGIN

		//dc.w 0xA9FF
		movem.l d0-d7/a0-a6,-(a7)
		moveq #0,d0
		moveq #0,d1
		moveq #0,d2
		moveq #0,d3
		move.w x1,d0
		move.w y1,d1
		move.w x2,d2
		move.w y2,d3
		sub.w d0,d2			// d2 = !x = ( x2 - x1 )
		bpl.s @noswap
		exg d1,d3			// swap endpoints
		neg.w d2			// !x = -!x = ( x1 - x2 )
		sub.w d2,d0			// d0 =  x1 - ( x1 - x2 ) = x2
@noswap:
		sub.w d1,d3			// d3 = !y = ( y2 - y1 )
		move.l baseAddr,a0
		move.l rowBytes,d6	
		mulu.l d6,d1		
		add.l d1,d0			
		add.l d0,a0			// a0 = address of the first point to draw
		move.b color,d7
// now we can freely use d0/d1 
		tst.w d3			// Line upward?
		bmi @LineUp			// yes
		beq @Horiz			// horizontal line
		tst.w d2			// !x = 0 - vertical line downward
		beq @VertDn
		cmp.w d2,d3			// test line slope
		beq @DiagDn			// slope = 1
		bpl @SteepDn		// slope > 1

		moveq #0,d0			// lower 32bits
		divu d2,d3:d0		// d0 = 32bit ( !y/!x )
		moveq #0,d1			// accumulated error
@FDLoop	move.b d7,(a0)+		// Plot a point and move 1 pixel rightward
		add.l d0,d1			// Add an accumulated error
		bcs.s @FDCarry
		dbra d2,@FDLoop		
		bra @exit	
@FDCarry
		add.l d6,a0			// One line down
		dbra d2,@FDLoop
		bra @exit	
@DiagDn
		addq.l #1,d6		// Offset is rowbytes+1
@DDLoop
		move.b d7,(a0)		// draw line
		add.l d6,a0
		dbra d2,@DDLoop
		bra @exit
@SteepDn
		moveq #0,d0
		divu d3,d2:d0		// d0 = 32bit ( !x/!y )
		moveq #0,d1			// accumulated error
@SDLoop
		move.b d7,(a0)		// plot a point
		add.l d6,a0			// move 1 pixel downward
		add.l d0,d1			// Add an accumulated error
		bcs.s @SDCarry
		dbra d3,@SDLoop
		bra @exit
@SDCarry
		addq.l #1,a0
		dbra d3,@SDLoop
		bra @exit
@Horiz
		move.b d7,(a0)+
		dbra d2,@Horiz
		bra @exit
@VertDn
		move.b d7,(a0)		// offset is rowbytes
		add.l d6,a0
		dbra d3,@VertDn
		bra @exit
@LineUp
		neg.w d3	
		tst.w d2
		beq @VertUp		
		cmp.w d2,d3
		beq @DiagUp			// slope = -1
		bpl @SteepUp		// slope < -1
		
		moveq #0,d0			// lower 32bits
		divu d2,d3:d0		// d0 = 32bit ( !y/!x )
		moveq #0,d1			// accumulated error
@FULoop	move.b d7,(a0)+		// Plot a point and move 1 pixel rightward
		add.l d0,d1			// Add an accumulated error
		bcs.s @FUCarry
		dbra d2,@FULoop		
		bra @exit	
@FUCarry
		sub.l d6,a0			// One line up
		dbra d2,@FULoop
		bra @exit	
@DiagUp
		subq.l #1,d6		// offset is 1-rowbytes
@DULoop
		move.b d7,(a0)		// draw line
		sub.l d6,a0
		dbra d2,@DULoop
		bra @exit
@SteepUp
		moveq #0,d0
		divu d3,d2:d0		// d0 = 32bit ( !x/!y )
		moveq #0,d1			// accumulated error
@SULoop
		move.b d7,(a0)		// plot a point
		sub.l d6,a0			// move 1 pixel upward
		add.l d0,d1			// Add an accumulated error
		bcs.s @SUCarry
		dbra d3,@SULoop
		bra @exit
@SUCarry
		addq.l #1,a0
		dbra d3,@SULoop
		bra @exit
@VertUp
		move.b d7,(a0)		// offset is -rowbytes
		sub.l d6,a0
		dbra d3,@VertUp
@exit:	movem.l (a7)+,d0-d7/a0-a6
	
	SW_ASM_END
	
	
	if (gSWmmuMode != true32b)
	{
		SwapMMUMode(&mmuMode);
	}
}

